import uuid
from asyncio import Future
from collections import defaultdict
from functools import partial
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, List, Literal, Optional, Tuple
from warnings import warn

import dash
import dash_mantine_components as dmc
from dash import Input, Output, State, ctx, dcc, html

from xiplot.tabs import Tab
from xiplot.utils.components import FlexRow
from xiplot.utils.layouts import layout_wrapper


class Plugins(Tab):
    @staticmethod
    def register_callbacks(app, df_from_store, df_to_store, plugin_dir=""):
        @app.callback(
            Output("plugin-tab-notify-container", "children"),
            Output("plugin_paths", "value"),
            Output("plugin-search-input", "value"),
            Output("plugin_paths", "options"),
            Output("plugin-tab-load-progress", "value"),
            Output("plugin-reload-button", "disabled"),
            Input("plugin-submit-button", "n_clicks"),
            Input("plugin-tab-load-success", "value"),
            Input("plugin-tab-load-exception", "value"),
            State("plugin_paths", "value"),
            State("plugin-search-input", "value"),
            State("plugin-tab-load-progress", "value"),
        )
        def load_plugin(
            plugin_btn,
            success,
            exception,
            plugin_path,
            plugin_search,
            progress_id,
        ):
            trigger = ctx.triggered_id

            if trigger == "plugin-tab-load-success":
                return (
                    [
                        dmc.Notification(
                            id=progress_id,
                            color="green",
                            title="Success",
                            message="The plugin was successfully installed!",
                            action="update",
                            autoClose=5000,
                        ),
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Information",
                            message=(
                                "You can now activate the newly installed"
                                " plugin by clicking the 'Reload χiplot'"
                                " button in the 'Plugins' tab."
                            ),
                            action="show",
                            autoClose=False,
                        ),
                    ],
                    dash.no_update,
                    "",
                    get_suggested_plugin_options(plugin_dir=plugin_dir),
                    "",
                    False,
                )
            if trigger == "plugin-tab-load-exception":
                exception = str(exception)
                exception = exception[exception.find("|") + 1 :]  # noqa: E203

                return (
                    dmc.Notification(
                        id=progress_id,
                        color="red",
                        title="Error",
                        message=exception,
                        action="update",
                        autoClose=False,
                    ),
                    dash.no_update,
                    "",
                    get_suggested_plugin_options(plugin_dir=plugin_dir),
                    "",
                    dash.no_update,
                )

            plugin_source = None

            if not plugin_path and plugin_search:
                plugin_source = plugin_search
                install_plugin = install_remote_plugin

            if plugin_path:
                plugin_source = plugin_path
                install_plugin = install_local_plugin

            if not plugin_source:
                return (
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message="You have not selected a plugin to install.",
                        action="show",
                        autoClose=10000,
                    ),
                    dash.no_update,
                    "",
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

            if progress_id:
                return (
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="yellow",
                        title="Warning",
                        message=(
                            "Please wait until the previous plugin has"
                            " finished installing."
                        ),
                        action="show",
                        autoClose=10000,
                    ),
                    dash.no_update,
                    "",
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

            try:
                install_plugin(plugin_source)

                progress_id = str(uuid.uuid4())

                return (
                    dmc.Notification(
                        id=progress_id,
                        color="blue",
                        title="Processing",
                        message="Installing the plugin",
                        action="show",
                        loading=True,
                        autoClose=False,
                        disallowClose=True,
                    ),
                    None,
                    "",
                    dash.no_update,
                    progress_id,
                    dash.no_update,
                )
            except Exception as err:
                return (
                    dmc.Notification(
                        id=str(uuid.uuid4()),
                        color="red",
                        title="Error",
                        message=(
                            f"Installing a plugin from '{plugin_source}'"
                            f" failed with the following error: {err}"
                        ),
                        action="show",
                        autoClose=False,
                    ),
                    None,
                    "",
                    dash.no_update,
                    dash.no_update,
                    dash.no_update,
                )

        @app.callback(
            Output("plugin-search-input", "value"),
            Input("plugin_paths", "search_value"),
        )
        def sync_search_input(search):
            if search == "":
                return dash.no_update
            return search

        @app.callback(
            Output("plugin-reload-modal", "opened"),
            Output("plugin-reload-activations", "children"),
            Input("plugin-reload-button", "n_clicks"),
            Input("plugin-reload-accept", "n_clicks"),
            Input("plugin-reload-cancel", "n_clicks"),
            State("plugin-reload-modal", "opened"),
            prevent_initial_call=True,
        )
        def update_plugin_reload_modal(reload, accept, cancel, opened):
            if ctx.triggered_id == "plugin-reload-accept":
                import pyodide

                get_plugins_cached.cache = dict()

                pyodide.code.run_js(
                    'self.postMessage({ jsEval: "window.web_dash.reload()" })'
                )

                return dash.no_update, dash.no_update

            plugins = defaultdict(set)

            for (
                kind,
                name,
                path,
                plugin,
            ) in get_all_loaded_plugins():
                plugins[path.split(":")[0].split(".")[0]].add(kind)

            for (
                kind,
                name,
                path,
                plugin,
            ) in get_all_loaded_plugins_cached():
                plugins[path.split(":")[0].split(".")[0]].discard(kind)

            new_plugins = []

            for plugin, kinds in plugins.items():
                if len(kinds) == 0:
                    continue

                new_plugins.append(
                    dmc.ListItem(f"{plugin}: {', '.join(kinds)}")
                )

            return not opened, new_plugins

    @staticmethod
    def create_layout(plugin_dir=""):
        uploader = dcc.Upload(
            id="plugin_uploader",
            children=html.Div(
                # [
                #     "Drag and Drop or ",
                #     html.A("Select a Python .whl File"),
                #     " to upload",
                # ]
                "Uploading new plugins is not yet supported."
            ),
            className="dcc-upload",
            disabled=True,
        )

        return FlexRow(
            html.Div(
                [
                    layout_wrapper(
                        component=FlexRow(
                            dcc.Dropdown(
                                get_suggested_plugin_options(
                                    plugin_dir=plugin_dir
                                ),
                                id="plugin_paths",
                                placeholder=(
                                    "Select a suggested plugin or type its URL"
                                    " or PyPi package name"
                                ),
                            ),
                            layout_wrapper(
                                component=dcc.Input(id="plugin-search-input"),
                                style={"display": "none"},
                            ),
                            html.Button(
                                "Install",
                                id="plugin-submit-button",
                                n_clicks=0,
                                className="button",
                            ),
                        ),
                        title="Install a Plugin",
                        css_class="dash-dropdown",
                    ),
                    html.Br(),
                    layout_wrapper(
                        component=FlexRow(
                            dcc.Dropdown(
                                get_loaded_plugin_options(),
                                id="loaded_plugins",
                                clearable=False,
                                searchable=False,
                                placeholder=(
                                    "Check the list of activated plugins"
                                ),
                            ),
                            html.Button(
                                "Reload χiplot",
                                id="plugin-reload-button",
                                n_clicks=0,
                                className="button",
                                disabled=True,
                            ),
                            dmc.Modal(
                                title="Do you really want to reload χiplot?",
                                id="plugin-reload-modal",
                                zIndex=10000,
                                children=[
                                    dmc.Text(
                                        "Please ensure that you have"
                                        " downloaded your plots and data so"
                                        " that nothing is lost. The following"
                                        " recently installed plugins will"
                                        " become activated after the reload:"
                                    ),
                                    dmc.Space(h=10),
                                    dmc.List(
                                        [],
                                        id="plugin-reload-activations",
                                        withPadding=True,
                                    ),
                                    dmc.Space(h=20),
                                    dmc.Group(
                                        [
                                            dmc.Button(
                                                "Reload",
                                                id="plugin-reload-accept",
                                            ),
                                            dmc.Button(
                                                "Cancel",
                                                color="red",
                                                variant="outline",
                                                id="plugin-reload-cancel",
                                            ),
                                        ],
                                        position="right",
                                    ),
                                ],
                            ),
                        ),
                        title="Active Plugins",
                        css_class="dash-dropdown",
                    ),
                ],
                className="stretch",
            ),
            html.Div(
                [uploader],
                id="plugin_uploader_container",
                className="dash-uploader",
            ),
            id="control_plugin_content-container",
        )

    @staticmethod
    def create_layout_globals():
        return html.Div(
            [
                html.Div(
                    id="plugin-tab-notify-container", style={"display": "none"}
                ),
                html.Div(
                    id="plugin-tab-upload-notify-container",
                    style={"display": "none"},
                ),
                dcc.Input(
                    id="plugin-tab-load-progress",
                    style={"display": "none"},
                    type="text",
                ),
                dcc.Input(
                    id="plugin-tab-load-success",
                    style={"display": "none"},
                    type="text",
                ),
                dcc.Input(
                    id="plugin-tab-load-exception",
                    style={"display": "none"},
                    type="text",
                ),
            ]
        )


def get_plugins(
    plugin_type: Literal["read", "write", "plot", "global", "callback"],
) -> List[Tuple[str, str, Any]]:
    """Get a list of all plugins of the specified type.

    Args:
        plugin_type: The type of xiplot plugin.

    Returns:
        A list of tuples of each plugin's name, path, and its
        loaded `entry_point`.
    """
    try:
        # Python 3.10+
        plugins = entry_points(group=f"xiplot.plugin.{plugin_type}")
    except TypeError:
        # Python 3.8-3.9
        plugins = entry_points().get(f"xiplot.plugin.{plugin_type}", ())

    loaded_plugins = []
    for plugin in plugins:
        try:
            loaded_plugins.append((plugin.name, plugin.value, plugin.load()))
        except Exception as e:
            warn(f"Could not load plugin {plugin}: {e}")

    return loaded_plugins


def get_plugins_cached(
    plugin_type: Literal["read", "write", "plot", "global", "callback"],
) -> List[Tuple[str, str, Any]]:
    """Get a list of all plugins of the specified type
     (this call is cached for future reuse).

    Args:
        plugin_type: The type of xiplot plugin.

    Returns:
        A list of tuples of each plugin's name, path, and its
        loaded `entry_point`.
    """
    if getattr(get_plugins_cached, "cache", None) is None:
        get_plugins_cached.cache = dict()

    cached_plugins_of_type = get_plugins_cached.cache.get(plugin_type, None)

    if cached_plugins_of_type is not None:
        return cached_plugins_of_type

    loaded_plugins = get_plugins(plugin_type)
    get_plugins_cached.cache[plugin_type] = loaded_plugins

    return loaded_plugins


def get_all_loaded_plugins() -> List[Tuple[str, str, str, Any]]:
    """Collects the list of all loaded plugins and their types.

    Returns:
        A list of tuples of each plugin's type, name, path, and its
        loaded `entry_point`.

    """
    plugins = []

    for plugin_type in ["read", "write", "plot", "global", "callback"]:
        for name, path, plugin in get_plugins(plugin_type):
            plugins.append((plugin_type, name, path, plugin))

    return plugins


def get_all_loaded_plugins_cached() -> List[Tuple[str, str, str, Any]]:
    """Collects the list of all cached loaded plugins and their types.

    Returns:
        A list of tuples of each plugin's type, name, path, and its
        loaded `entry_point`.

    """
    plugins = []

    for plugin_type in ["read", "write", "plot", "global", "callback"]:
        for name, path, plugin in get_plugins_cached(plugin_type):
            plugins.append((plugin_type, name, path, plugin))

    return plugins


def is_dynamic_plugin_loading_supported() -> bool:
    """Checks if xiplot supports dynamically loading new plugins at runtime.

    Returns:
        `True` if dynamic loading is supported, `False` otherwise.
    """
    try:
        import micropip  # noqa: F401
    except ImportError:
        return False

    return True


def get_loaded_plugin_options():
    plugins = defaultdict(set)

    for (
        kind,
        name,
        path,
        plugin,
    ) in get_all_loaded_plugins_cached():
        plugins[path.split(":")[0].split(".")[0]].add(kind)

    plugin_options = []

    for plugin, kinds in plugins.items():
        if len(kinds) > 0:
            kinds = f": {', '.join(kinds)}"
        else:
            kinds = ""

        plugin_options.append(
            {
                "label": f"{plugin}{kinds}",
                "value": plugin,
                "disabled": True,
            }
        )

    return plugin_options


def get_suggested_plugin_options(plugin_dir=""):
    return [
        {"label": fp.name, "value": str(fp)}
        for fp in get_plugin_filepaths(plugin_dir=plugin_dir)
    ]


def get_plugin_filepaths(plugin_dir=""):
    try:
        return sorted(
            (
                fp
                for fp in Path(plugin_dir).iterdir()
                if fp.is_file() and fp.suffix == ".whl"
            ),
            reverse=True,
        )

    except FileNotFoundError:
        return []


def install_local_plugin(plugin_path: str):
    import asyncio

    try:
        import micropip
    except ImportError:
        raise NotImplementedError(
            "Installing new plugins is only supported in WASM"
        )

    plugin_path = Path(plugin_path).resolve()

    asyncio.ensure_future(
        micropip.install(f"emfs:{str(plugin_path)}")
    ).add_done_callback(
        partial(__micropip_install_callback, try_remove_file=plugin_path)
    )


def install_remote_plugin(plugin_source: str):
    import asyncio

    try:
        import micropip
    except ImportError:
        raise NotImplementedError(
            "Installing new plugins is only supported in WASM"
        )

    # Protect against arbitrary local installs
    if plugin_source.startswith("emfs:"):
        raise ValueError("Plugin source must be a URL or PyPi package name")

    asyncio.ensure_future(micropip.install(plugin_source)).add_done_callback(
        partial(__micropip_install_callback, try_remove_file=None)
    )


def __micropip_install_callback(
    future: Future, try_remove_file: Optional[Path]
):
    import pyodide

    result = str(uuid.uuid4())

    try:
        future.result()
        target = "success"
    except Exception as err:
        target = "exception"
        result += f"|{err}"

    if target == "success" and try_remove_file is not None:
        try:
            try_remove_file.unlink()
        except Exception:
            pass

    js_callback_code = f"""
window.Object.getOwnPropertyDescriptor(
    window.HTMLInputElement.prototype, "value",
).set.call(
    window.document.getElementById("plugin-tab-load-{target}"), {repr(result)},
);
window.document.getElementById("plugin-tab-load-{target}").dispatchEvent(
    new Event("input", {{ bubbles: true }}),
);
    """

    pyodide.code.run_js(
        f"self.postMessage({{ jsEval: {repr(js_callback_code)} }})"
    )



import dash_mantine_components as dmc
from dash import dcc, html

from xiplot.tabs.cluster import Cluster
from xiplot.tabs.data import Data
from xiplot.tabs.embedding import Embedding
from xiplot.tabs.plots import Plots
from xiplot.tabs.plugins import (
    Plugins,
    get_plugins_cached,
    is_dynamic_plugin_loading_supported,
)
from xiplot.tabs.settings import Settings
from xiplot.utils.components import ClusterDropdown


class XiPlot:
    def __init__(
        self, app, df_from_store, df_to_store, data_dir="", plugin_dir=""
    ) -> None:
        self.app = app
        self.app.title = "χiplot"

        try:
            import dash_uploader as du

            du.configure_upload(
                app=self.app, folder="uploads", use_upload_id=False
            )
        except (ImportError, AttributeError):
            pass

        TABS = [Data, Plots, Cluster, Embedding, Settings]

        if is_dynamic_plugin_loading_supported():
            TABS.insert(-1, Plugins)

        self.app.layout = dmc.NotificationsProvider(
            html.Div(
                [
                    html.Div(
                        [
                            app_links(),
                            app_logo(),
                            dcc.Tabs(
                                [
                                    dcc.Tab(
                                        [
                                            (
                                                t.create_layout(
                                                    data_dir=data_dir
                                                )
                                                if t == Data
                                                else (
                                                    t.create_layout(
                                                        plugin_dir=plugin_dir
                                                    )
                                                    if t == Plugins
                                                    else t.create_layout()
                                                )
                                            )
                                        ],
                                        label=t.name(),
                                        value=(
                                            f"control-{t.name().lower()}-tab"
                                        ),
                                    )
                                    for t in TABS
                                ],
                                id="control-tabs",
                                value=f"control-{TABS[0].name().lower()}-tab",
                            ),
                        ],
                        id="control",
                        className="control",
                    ),
                    html.Div(id="plots"),
                    dcc.Store(id="data_frame_store"),
                    dcc.Store(id="auxiliary_store"),
                    dcc.Store(id="metadata_store"),
                    dcc.Store(id="lastly_clicked_point_store"),
                    dcc.Store(id="lastly_hovered_point_store"),
                    html.Div(
                        [t.create_layout_globals() for t in TABS],
                        id="globals",
                    ),
                    html.Div(
                        [g() for (_, _, g) in get_plugins_cached("global")],
                        id="plugin-globals",
                    ),
                ],
                id="main",
            ),
            position="top-right",
        )

        for tab in TABS:
            (
                tab.register_callbacks(
                    app, df_from_store, df_to_store, data_dir=data_dir
                )
                if tab == Data
                else (
                    tab.register_callbacks(
                        app,
                        df_from_store,
                        df_to_store,
                        plugin_dir=plugin_dir,
                    )
                    if tab == Plugins
                    else tab.register_callbacks(
                        app, df_from_store, df_to_store
                    )
                )
            )

        for _, _, cb in get_plugins_cached("callback"):
            cb(app, df_from_store, df_to_store)

        ClusterDropdown.register_callbacks(app)


def app_logo():
    return html.Div([html.H1("χiplot")], id="logo", className="logo")


def app_links():
    return html.Div(
        [
            html.A(
                html.Img(src="assets/book.svg", alt="Documentation"),
                href="https://github.com/edahelsinki/xiplot/blob/main/docs/README.md",
                title="Documentation",
                target="_blank",
            ),
            html.A(
                html.Img(src="assets/github.svg", alt="GitHub"),
                href="https://github.com/edahelsinki/xiplot",
                title="GitHub",
                target="_blank",
            ),
        ],
        className="links",
    )



import dash_extensions.enrich as enrich
import pandas as pd
from dash_extensions.enrich import (
    CycleBreakerTransform,
    DashProxy,
    MultiplexerTransform,
    ServersideOutputTransform,
)

from xiplot.app import XiPlot
from xiplot.utils.store import ServerSideStoreBackend


def setup_xiplot_dash_app(
    unsafe_local_server=False, data_dir="", plugin_dir="", **kwargs
):
    dash_transforms = [
        MultiplexerTransform(),
        CycleBreakerTransform(),
    ]

    if unsafe_local_server:
        dash_transforms.append(
            ServersideOutputTransform(
                backend=ServerSideStoreBackend(),
                session_check=False,
                arg_check=False,
            )
        )

        def df_from_store(df):
            if isinstance(df, pd.DataFrame):
                return df.copy(deep=False)
            return df

        def df_to_store(df):
            return df

    else:

        def df_from_store(df):
            return pd.read_json(df, orient="split")

        def df_to_store(df):
            return df.to_json(date_format="iso", orient="split")

    dash = DashProxy(
        "xiplot.app",
        suppress_callback_exceptions=True,
        transforms=dash_transforms,
        prevent_initial_callbacks=True,
        **kwargs,
    )

    _app = XiPlot(  # noqa: F841
        app=dash,
        df_from_store=df_from_store,
        df_to_store=df_to_store,
        data_dir=data_dir,
        plugin_dir=plugin_dir,
    )

    return dash


""" Monkey patch for dash_extensions to remove print """


# https://github.com/thedirtyfew/dash-extensions/commit/d843b63
def _new_get_cache_id(func, output, args, session_check=None, arg_check=True):
    all_args = [func.__name__, enrich._create_callback_id(output)]
    if arg_check:
        all_args += list(args)
    if session_check:
        all_args += [enrich._get_session_id()]
    return enrich.hashlib.md5(enrich.json.dumps(all_args).encode()).hexdigest()


enrich._get_cache_id = _new_get_cache_id



import pandas as pd
import plotly.express as px
from dash import ALL, MATCH, Input, Output, ctx, dcc
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import ID_HOVERED
from xiplot.utils.auxiliary import (
    SELECTED_COLUMN_NAME,
    decode_aux,
    get_clusters,
    merge_df_aux,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ClusterDropdown,
    ColumnDropdown,
    PdfButton,
    PlotData,
)
from xiplot.utils.dataframe import get_numeric_columns
from xiplot.utils.layouts import layout_wrapper


class Histogram(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), {"type": "histogram"})

        @app.callback(
            Output({"type": "histogram", "index": MATCH}, "figure"),
            Input(cls.get_id(MATCH, "x_axis_dropdown"), "value"),
            Input(ClusterDropdown.get_id(MATCH), "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def tmp(x_axis, selected_clusters, hover, df, aux, template=None):
            # Try branch for testing
            try:
                if ctx.triggered_id == "data_frame_store":
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                pass

            return Histogram.render(
                x_axis,
                selected_clusters,
                hover,
                df_from_store(df),
                decode_aux(aux),
                template,
            )

        PlotData.register_callback(
            cls.name(),
            app,
            (
                Input(cls.get_id(ALL, "x_axis_dropdown"), "value"),
                Input(ClusterDropdown.get_id(ALL), "value"),
            ),
            lambda i: dict(axes=dict(x=i[0]), classes=i[1] or []),
        )

        ColumnDropdown.register_callback(
            app,
            cls.get_id(ALL, "x_axis_dropdown"),
            df_from_store,
            numeric=True,
        )

        return [tmp]

    @staticmethod
    def render(x_axis, selected_clusters, hover, df, aux, template=None):
        df = merge_df_aux(df, aux)
        clusters = get_clusters(aux, df.shape[0])
        if type(selected_clusters) == str:
            selected_clusters = [selected_clusters]
        if not selected_clusters:
            selected_clusters = clusters.categories

        props_dict = {"all": []}
        for s in selected_clusters:
            if s != "all":
                props_dict[s] = []

        for c, p in zip(clusters, df[x_axis]):
            if c != "all" and c in selected_clusters:
                props_dict[c].append(p)
            props_dict["all"].append(p)

        clusters_col = []
        x = []
        for s in selected_clusters:
            clusters_col += [s for _ in props_dict[s]]
            x += props_dict[s]

        dff = pd.DataFrame({"Clusters": clusters_col, x_axis: x})

        fig_property = px.histogram(
            dff,
            x=x_axis,
            color="Clusters",
            hover_data={
                "Clusters": False,
                x_axis: False,
            },
            color_discrete_map=cluster_colours(),
            opacity=0.5,
            histnorm="probability density",
            template=template,
        )

        fig_property.update_layout(
            hovermode="x unified",
            showlegend=False,
            barmode="overlay",
            yaxis=dict(
                tickformat=".2%",
            ),
        )

        if hover is not None:
            fig_property.add_vline(
                df[x_axis][hover],
                line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                layer="below",
            )
        if SELECTED_COLUMN_NAME in aux:
            color = "#DDD" if template and "dark" in template else "#333"
            for x in df[x_axis][aux[SELECTED_COLUMN_NAME]]:
                fig_property.add_vline(
                    x, line=dict(color=color, width=0.5), layer="below"
                )

        return fig_property

    @classmethod
    def create_layout(cls, index, df, columns, config=dict()):
        import jsonschema

        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object",
                properties=dict(
                    axes=dict(
                        type="object", properties=dict(x=dict(type="string"))
                    ),
                    classes=dict(
                        type="array",
                        items=dict(enum=list(cluster_colours().keys())),
                        uniqueItems=True,
                    ),
                ),
            ),
        )
        num_columns = get_numeric_columns(df, columns)

        try:
            x_axis = config["axes"]["x"]
        except Exception:
            x_axis = None

        if x_axis is None and len(num_columns) > 0:
            x_axis = num_columns[0]

        if x_axis is None:
            raise Exception("The dataframe contains no numeric columns")

        classes = config.get("classes", [])

        return [
            dcc.Graph(id={"type": "histogram", "index": index}),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "x_axis_dropdown"),
                    value=x_axis,
                    clearable=False,
                    options=[x_axis],
                ),
                css_class="dd-single",
                title="x axis",
            ),
            layout_wrapper(
                component=ClusterDropdown(
                    index=index, multi=True, value=classes, clearable=True
                ),
                title="Cluster Comparison",
                css_class="dd-single",
            ),
        ]



import plotly.express as px


def cluster_colours():
    return {
        "all": px.colors.qualitative.Vivid[-1],
        **{
            f"c{i+1}": c
            for i, c in enumerate(px.colors.qualitative.Vivid[:-1])
        },
    }


def KMeans(n_clusters: int = 8, **kwargs):
    """A wrapper around `sklearn.cluster.KMeans` that changes `n_init="warn"`
    to `n_init="auto"`. This is needed to avoid a warning for scikit-learn
    versions `>=1.2,<1.4`. Older and newer versions are not affected (unless
    `n_init="warn"` is manually specified).

    NOTE: This function lazily loads scikit-learn (so the first call might be slow).

    Args:
        n_clusters: The number of clusters. Defaults to 8.
        **kwargs: See `sklearn.cluster.KMeans`.

    Returns:
        An instance of `sklearn.cluster.KMeans`.
    """
    from sklearn.cluster import KMeans

    km = KMeans(n_clusters, **kwargs)
    if km.n_init == "warn":
        km.n_init = "auto"
    return km



import uuid
from collections import defaultdict
from itertools import product

import dash
import dash_mantine_components as dmc
import numpy as np
import pandas as pd
import plotly.express as px
from dash import MATCH, Input, Output, ctx, dcc, html
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.plugin import ID_HOVERED
from xiplot.utils.auxiliary import (
    SELECTED_COLUMN_NAME,
    decode_aux,
    get_clusters,
    merge_df_aux,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import (
    ClusterDropdown,
    ColumnDropdown,
    PdfButton,
    PlotData,
)
from xiplot.utils.layouts import layout_wrapper


class Barplot(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), {"type": "barplot"})

        @app.callback(
            Output({"type": "barplot", "index": MATCH}, "figure"),
            Output(
                {"type": "barplot-notify-container", "index": MATCH},
                "children",
            ),
            Input(cls.get_id(MATCH, "x_axis_dropdown"), "value"),
            Input(cls.get_id(MATCH, "y_axis_dropdown"), "value"),
            Input(ClusterDropdown.get_id(MATCH), "value"),
            Input({"type": "order_dropdown", "index": MATCH}, "value"),
            Input(ID_HOVERED, "data"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def tmp(
            x_axis,
            y_axis,
            selected_clusters,
            order,
            hover,
            df,
            aux,
            template=None,
        ):
            try:
                if ctx.triggered_id == "data_frame_store":
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                pass

            try:
                return (
                    Barplot.render(
                        x_axis,
                        y_axis,
                        selected_clusters,
                        order,
                        hover,
                        df_from_store(df),
                        decode_aux(aux),
                        template,
                    ),
                    dash.no_update,
                )
            except Exception as err:
                return dash.no_update, dmc.Notification(
                    id=str(uuid.uuid4()),
                    color="yellow",
                    title="Warning",
                    message=f"Barplot error: {err}",
                    action="show",
                    autoClose=10000,
                )

        PlotData.register_callback(
            cls.name(),
            app,
            [
                Input(cls.get_id(MATCH, "x_axis_dropdown"), "value"),
                Input(cls.get_id(MATCH, "y_axis_dropdown"), "value"),
                Input(ClusterDropdown.get_id(MATCH), "value"),
                Input({"type": "order_dropdown", "index": MATCH}, "value"),
            ],
            lambda i: dict(
                axes=dict(x=i[0], y=i[1]),
                classes=i[2] or [],
                order=i[3],
            ),
        )

        ColumnDropdown.register_callback(
            app,
            cls.get_id(MATCH, "x_axis_dropdown"),
            df_from_store,
            category=True,
        )
        ColumnDropdown.register_callback(
            app,
            cls.get_id(MATCH, "y_axis_dropdown"),
            df_from_store,
            options=["frequency"],
            numeric=True,
        )

        return [tmp]

    @staticmethod
    def render(
        x_axis,
        y_axis,
        selected_clusters,
        order,
        hover,
        df,
        aux,
        template=None,
    ):
        df = merge_df_aux(df, aux)
        if "frequency" not in df.columns:
            df["frequency"] = [1 for _ in range(len(df))]

        if x_axis == y_axis:
            raise Exception("The x and y axis must be different")

        clusters = get_clusters(aux, df.shape[0])
        if not selected_clusters:
            selected_clusters = clusters.categories
        selected_clusters = set(selected_clusters)

        Xs = []
        Ys = []
        Cs = []

        if type(df[x_axis][0]) in [np.ndarray, list]:
            if y_axis == "frequency":
                df["frequency"] = [
                    f / max(len(xs), 1)
                    for f, xs in zip(df["frequency"], df[x_axis])
                ]

            for xs, y, c in zip(df[x_axis], df[y_axis], clusters):
                for x in xs:
                    if c in selected_clusters:
                        Xs.append(x)
                        Ys.append(y)
                        Cs.append(c)

                    if c != "all" and "all" in selected_clusters:
                        Xs.append(x)
                        Ys.append(y)
                        Cs.append("all")
        else:
            for x, y, c in zip(df[x_axis], df[y_axis], clusters):
                if c in selected_clusters:
                    Xs.append(x)
                    Ys.append(y)
                    Cs.append(c)

                if c != "all" and "all" in selected_clusters:
                    Xs.append(x)
                    Ys.append(y)
                    Cs.append("all")

        flat_df = pd.DataFrame(
            {
                x_axis: Xs,
                y_axis: Ys,
                **{
                    "Clusters": Cs
                    for _ in ((),)
                    if x_axis != "Clusters" and y_axis != "Clusters"
                },
            }
        )

        grouping = flat_df.groupby([x_axis, "Clusters"])[y_axis]

        if y_axis == "frequency" and type(df[x_axis][0]) not in [
            np.ndarray,
            list,
        ]:
            dff = grouping.sum().to_frame().reset_index()
        else:
            dff = grouping.mean().to_frame().reset_index()
            dff["Error"] = grouping.sem().values

        if order == "total" or len(selected_clusters) <= 1:
            order_lookup = defaultdict(lambda: 0.0)

            for x, y in zip(dff[x_axis].values, dff[y_axis].values):
                order_lookup[x] += abs(y)
        elif order == "reldiff":
            value_lookup = defaultdict(list)

            if "Error" in dff.columns:
                for x, y, e in zip(
                    dff[x_axis].values, dff[y_axis].values, dff["Error"].values
                ):
                    value_lookup[x].append((y, e))
            else:
                for x, y in zip(dff[x_axis].values, dff[y_axis].values):
                    value_lookup[x].append((y, 1))

            order_lookup = dict()

            for x, vs in value_lookup.items():
                if len(vs) <= 1:
                    order_lookup[x] = 0.0
                else:
                    maxdiff = 0.0

                    for (v1, e1), (v2, e2) in product(vs, vs):
                        if e1 == 0.0 and e2 == 0.0:
                            e1 = 0.001

                        diff = abs(v1 - v2) / np.sqrt(e1 * e1 + e2 * e2)

                        maxdiff = max(maxdiff, diff)

                    order_lookup[x] = maxdiff

        top_bars = sorted(
            order_lookup.keys(), key=lambda x: order_lookup[x], reverse=True
        )[:10]

        dff.drop(
            index=[
                i for i, x in zip(dff.index, dff[x_axis]) if x not in top_bars
            ],
            inplace=True,
        )

        fig = px.bar(
            dff,
            x_axis,
            y_axis,
            error_y="Error" if "Error" in dff.columns else None,
            color="Clusters",
            barmode="group",
            hover_data={
                "Clusters": False,
                x_axis: False,
            },
            color_discrete_map=cluster_colours(),
            template=template,
        )

        fig.update_layout(
            hovermode="x unified",
            showlegend=False,
            xaxis=dict(
                fixedrange=True, categoryorder="array", categoryarray=top_bars
            ),
            yaxis=dict(fixedrange=True),
        )

        if y_axis != "frequency":
            if hover is not None:
                fig.add_hline(
                    df[y_axis][hover],
                    line=dict(color="rgba(0.5,0.5,0.5,0.5)", dash="dash"),
                    layer="below",
                )

            if SELECTED_COLUMN_NAME in aux:
                color = "#DDD" if template and "dark" in template else "#333"
                for x in df[y_axis][aux[SELECTED_COLUMN_NAME]]:
                    fig.add_hline(
                        x, line=dict(color=color, width=0.5), layer="below"
                    )

        return fig

    @classmethod
    def create_layout(cls, index, df, columns, config=dict()):
        import jsonschema

        x_columns = ColumnDropdown.get_columns(
            df, pd.DataFrame(), category=True
        )
        y_columns = ColumnDropdown.get_columns(
            df, pd.DataFrame(), ["frequency"], numeric=True
        )
        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object",
                properties=dict(
                    axes=dict(
                        type="object",
                        properties=dict(
                            x=dict(type="string"), y=dict(type="string")
                        ),
                    ),
                    classes=dict(
                        type="array",
                        items=dict(enum=list(cluster_colours().keys())),
                        uniqueItems=True,
                    ),
                    order=dict(enum=["reldiff", "total"]),
                ),
            ),
        )

        try:
            x_axis = config["axes"]["x"]
        except Exception:
            x_axis = None

        if x_axis is None and len(x_columns) > 0:
            x_axis = x_columns[0]

        try:
            y_axis = config["axes"]["y"]
        except Exception:
            y_axis = "frequency"

        if x_axis == y_axis:
            raise Exception("The x and y axis must be different")

        classes = config.get("classes", [])
        order = config.get("order", "reldiff")

        return [
            dcc.Graph(id={"type": "barplot", "index": index}),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "x_axis_dropdown"),
                    value=x_axis,
                    clearable=False,
                    options=x_columns,
                ),
                css_class="dd-double-left",
                title="x axis",
            ),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "y_axis_dropdown"),
                    value=y_axis,
                    clearable=False,
                    options=y_columns,
                ),
                css_class="dd-double-right",
                title="y axis",
            ),
            layout_wrapper(
                component=ClusterDropdown(
                    index=index, multi=True, value=classes, clearable=True
                ),
                title="Cluster Comparison",
                css_class="dd-single cluster-comparison",
            ),
            layout_wrapper(
                component=dcc.Dropdown(
                    id={"type": "order_dropdown", "index": index},
                    value=order,
                    clearable=False,
                    options=["reldiff", "total"],
                ),
                css_class="dd-single",
                title="Comparison Order",
            ),
            html.Div(
                id={"type": "barplot-notify-container", "index": index},
                style={"display": "none"},
            ),
        ]



import json

import dash
import numpy as np
import pandas as pd
import plotly.express as px
from dash import ALL, MATCH, Input, Output, State, ctx, dcc
from dash.exceptions import PreventUpdate

from xiplot.plots import APlot
from xiplot.utils.auxiliary import (
    CLUSTER_COLUMN_NAME,
    SELECTED_COLUMN_NAME,
    decode_aux,
    encode_aux,
    get_clusters,
    get_selected,
    merge_df_aux,
)
from xiplot.utils.cluster import cluster_colours
from xiplot.utils.components import ColumnDropdown, PdfButton, PlotData
from xiplot.utils.dataframe import get_default_column, get_numeric_columns
from xiplot.utils.layouts import layout_wrapper
from xiplot.utils.scatterplot import get_row


class Scatterplot(APlot):
    @classmethod
    def register_callbacks(cls, app, df_from_store, df_to_store):
        PdfButton.register_callback(app, cls.name(), {"type": "scatterplot"})

        @app.callback(
            Output({"type": "scatterplot", "index": MATCH}, "figure"),
            Input(cls.get_id(MATCH, "x_axis_dropdown"), "value"),
            Input(cls.get_id(MATCH, "y_axis_dropdown"), "value"),
            Input(cls.get_id(MATCH, "color_dropdown"), "value"),
            Input(cls.get_id(MATCH, "symbol_dropdown"), "value"),
            Input({"type": "jitter-slider", "index": MATCH}, "value"),
            Input("data_frame_store", "data"),
            Input("auxiliary_store", "data"),
            Input("plotly-template", "data"),
            prevent_initial_call=False,
        )
        def tmp(
            x_axis,
            y_axis,
            color,
            symbol,
            jitter,
            df,
            aux,
            template=None,
        ):
            # Try branch for testing
            try:
                if ctx.triggered_id == "data_frame_store":
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                pass

            fig = Scatterplot.render(
                df_from_store(df),
                decode_aux(aux),
                x_axis,
                y_axis,
                color,
                symbol,
                jitter,
                template,
            )

            if fig is None:
                return dash.no_update

            return fig

        @app.callback(
            Output({"type": "jitter-slider", "index": MATCH}, "max"),
            Input({"type": "x_axis_dropdown", "index": MATCH}, "value"),
            State("data_frame_store", "data"),
        )
        def update_jitter_max(x_axis, df):
            df = df_from_store(df)
            if x_axis in df.columns.to_list():
                return (df[x_axis].max() - df[x_axis].min()) * 0.05
            return dash.no_update

        @app.callback(
            output=dict(
                aux=Output("auxiliary_store", "data"),
                click_store=Output("lastly_clicked_point_store", "data"),
                scatter=Output(
                    {"type": "scatterplot", "index": ALL}, "clickData"
                ),
            ),
            inputs=[
                Input({"type": "scatterplot", "index": ALL}, "clickData"),
                State("auxiliary_store", "data"),
            ],
        )
        def handle_click_events(click, aux):
            # Try branch for testing
            try:
                if ctx.triggered_id is None:
                    raise PreventUpdate()
            except PreventUpdate:
                raise
            except Exception:
                pass

            row = get_row(click)
            if row is None:
                raise PreventUpdate()

            if aux is None:
                return dash.no_update
            aux = decode_aux(aux)
            selected = get_selected(aux)
            selected[row] = not selected[row]
            aux[SELECTED_COLUMN_NAME] = selected

            return dict(
                aux=encode_aux(aux),
                click_store=row,
                scatter=[None] * len(click),
            )

        @app.callback(
            output=dict(
                hover_store=Output("lastly_hovered_point_store", "data"),
                scatter=Output(
                    {"type": "scatterplot", "index": ALL}, "hoverData"
                ),
            ),
            inputs=[
                Input({"type": "scatterplot", "index": ALL}, "hoverData"),
            ],
        )
        def handle_hover_events(hover):
            row = get_row(hover)
            if row is None:
                raise PreventUpdate()

            scatter_amount = len(hover)
            return dict(
                hover_store=row,
                scatter=[None] * scatter_amount,
            )

        @app.callback(
            Output("auxiliary_store", "data"),
            Input({"type": "scatterplot", "index": ALL}, "selectedData"),
            State("auxiliary_store", "data"),
            State("selection_cluster_dropdown", "value"),
            State("cluster_selection_mode", "value"),
        )
        def handle_cluster_drawing(
            selected_data, aux, cluster_id, selection_mode
        ):
            if not selected_data:
                return dash.no_update

            updated = False

            aux = decode_aux(aux)
            clusters = get_clusters(aux)
            if not selection_mode:
                clusters = clusters.set_categories(["c1", "c2"])
                clusters[:] = "c2"
            else:
                if cluster_id not in clusters.categories:
                    clusters = clusters.add_categories([cluster_id])

            for trigger in selected_data:
                if not trigger or not trigger["points"]:
                    continue

                updated = updated or len(trigger["points"]) > 0

                try:
                    if selection_mode:
                        for p in trigger["points"]:
                            clusters[p["customdata"][0]["index"]] = cluster_id
                    else:
                        for p in trigger["points"]:
                            clusters[p["customdata"][0]["index"]] = "c1"
                except Exception:
                    return dash.no_update

            if not updated:
                return dash.no_update

            aux[CLUSTER_COLUMN_NAME] = clusters
            return encode_aux(aux)

        PlotData.register_callback(
            cls.name(),
            app,
            (
                Input(cls.get_id(MATCH, "x_axis_dropdown"), "value"),
                Input(cls.get_id(MATCH, "y_axis_dropdown"), "value"),
                Input(cls.get_id(MATCH, "color_dropdown"), "value"),
                Input(cls.get_id(MATCH, "symbol_dropdown"), "value"),
                Input({"type": "jitter-slider", "index": MATCH}, "value"),
            ),
            lambda i: dict(
                axes=dict(x=i[0], y=i[1]),
                colour=i[2],
                symbol=i[3],
                jitter=i[4],
            ),
        )

        ColumnDropdown.register_callback(
            app,
            cls.get_id(ALL, "x_axis_dropdown"),
            df_from_store,
            numeric=True,
        )
        ColumnDropdown.register_callback(
            app,
            cls.get_id(ALL, "y_axis_dropdown"),
            df_from_store,
            numeric=True,
        )
        ColumnDropdown.register_callback(
            app,
            cls.get_id(ALL, "color_dropdown"),
            df_from_store,
        )
        ColumnDropdown.register_callback(
            app,
            cls.get_id(ALL, "symbol_dropdown"),
            df_from_store,
            category=True,
        )

        return [
            tmp,
            handle_click_events,
            handle_hover_events,
            handle_cluster_drawing,
        ]

    @staticmethod
    def render(
        df,
        aux,
        x_axis,
        y_axis,
        color=None,
        symbol=None,
        jitter=None,
        template=None,
    ):
        df = merge_df_aux(df, aux)
        if jitter:
            jitter = float(jitter)
        if type(jitter) == float:
            if jitter > 0:
                Z = df[[x_axis, y_axis]].to_numpy("float64")
                Z = np.random.normal(Z, jitter)
                jitter_df = pd.DataFrame(Z, columns=[x_axis, y_axis])

                # Distinguish axes' names if identical
                if x_axis == y_axis:
                    jitter_df.columns = [x_axis + "(1)", x_axis + "(2)"]

                df[["jitter-x", "jitter-y"]] = jitter_df

                # Set jitter results to the axes but keep the title of the axes
                x_title, y_title = x_axis, y_axis
                x_axis, y_axis = "jitter-x", "jitter-y"

        sizes = [0.5] * df.shape[0]
        if color and color in df:
            colors = df[color].copy()
        else:
            colors = [""] * df.shape[0]
        if SELECTED_COLUMN_NAME in aux:
            for id in np.where(aux[SELECTED_COLUMN_NAME])[0]:
                if isinstance(colors, pd.Series):
                    colors = pd.Categorical(colors).add_categories("*")
                sizes[id] = 5
                colors[id] = "*"

        df["__Sizes__"] = sizes
        df["__Color__"] = colors
        df["__Auxiliary__"] = [{"index": i} for i in range(len(df))]

        fig = px.scatter(
            data_frame=df,
            x=x_axis,
            y=y_axis,
            color="__Color__",
            symbol=symbol if symbol in df else None,
            size="__Sizes__" if 5 in sizes else None,
            opacity=1,
            color_discrete_map={
                "*": "#DDD" if template and "dark" in template else "#333",
                **cluster_colours(),
            },
            custom_data=["__Auxiliary__"],
            hover_data={"__Color__": False, "__Sizes__": False},
            render_mode="webgl",
            template=template,
        )
        fig.update_layout(
            showlegend=False, uirevision=json.dumps([x_axis, y_axis])
        )
        fig.update(layout_coloraxis_showscale=False)
        fig.update_traces(marker={"line": {"width": 0}})

        if jitter:
            fig.update_xaxes(title=x_title)
            fig.update_yaxes(title=y_title)

        return fig

    @classmethod
    def create_layout(cls, index, df, columns, config=dict()):
        import jsonschema

        jsonschema.validate(
            instance=config,
            schema=dict(
                type="object",
                properties=dict(
                    axes=dict(
                        type="object",
                        properties=dict(
                            x=dict(type="string"), y=dict(type="string")
                        ),
                    ),
                    colour=dict(type="string"),
                    symbol=dict(type="string"),
                    jitter=dict(type="number", minimum=0.0),
                ),
            ),
        )
        num_columns = get_numeric_columns(df, columns)

        axes = config.get("axes", dict())
        x_axis = axes.get("x", get_default_column(num_columns, "x"))
        y_axis = axes.get("y", get_default_column(num_columns, "y"))
        if x_axis is None or y_axis is None:
            raise Exception("The dataframe contains no numeric columns")

        scatter_colour = config.get("colour", CLUSTER_COLUMN_NAME)
        scatter_symbol = config.get("symbol", CLUSTER_COLUMN_NAME)
        jitter_slider = config.get("jitter", 0.0)

        df["__Auxiliary__"] = [{"index": i} for i in range(len(df))]

        return [
            dcc.Graph(id={"type": "scatterplot", "index": index}),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "x_axis_dropdown"),
                    options=num_columns,
                    value=x_axis,
                    clearable=False,
                ),
                css_class="dd-double-left",
                title="x",
            ),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "y_axis_dropdown"),
                    options=num_columns,
                    value=y_axis,
                    clearable=False,
                ),
                css_class="dd-double-right",
                title="y",
            ),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "color_dropdown"),
                    options=columns,
                    value=scatter_colour,
                    clearable=False,
                ),
                css_class="dd-double-left",
                title="target (color)",
            ),
            layout_wrapper(
                component=ColumnDropdown(
                    cls.get_id(index, "symbol_dropdown"),
                    options=columns,
                    value=scatter_symbol,
                ),
                css_class="dd-double-right",
                title="target (symbol)",
            ),
            layout_wrapper(
                component=dcc.Slider(
                    id={"type": "jitter-slider", "index": index},
                    min=0,
                    max=1,
                    value=jitter_slider,
                    marks=None,
                    tooltip={"placement": "bottom", "always_visible": True},
                ),
                css_class="slider-single",
                title="jitter",
            ),
        ]



import dash
import pandas as pd
from selenium.webdriver.common.by import By

from tests.util_test import (
    click_pdf_button,
    render_plot,
    select_dropdown,
    start_server,
)
from xiplot.plots.lineplot import Lineplot
from xiplot.utils.auxiliary import get_selected

(
    render,
    handle_hover_events,
    handle_click_events,
) = Lineplot.register_callbacks(dash.Dash(__name__), lambda x: x, lambda x: x)


def test_lineplot_browser(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Lineplot")

    # Check render
    plot = driver.find_element(By.CLASS_NAME, "dash-graph")
    assert "Lineplot" in plot.get_attribute("id")
    dropdowns = driver.find_element(By.ID, "plots").find_elements(
        By.CLASS_NAME, "dash-dropdown"
    )
    assert len(dropdowns) == 6

    # Change axis
    select_dropdown(dropdowns[0], "mpg")
    assert "mpg" in driver.find_element(By.CLASS_NAME, "xtitle").text

    # Change color
    select_dropdown(dropdowns[4], "cylinders")
    assert "cylinders" in driver.find_element(By.CLASS_NAME, "dash-graph").text

    # Download pdf
    click_pdf_button(driver)

    # Close browser
    assert dash_duo.get_logs() == [], "browser console should contain no error"
    driver.close()


def test_lineplot_create():
    fig = render(
        "col1",
        "col2",
        "Clusters",
        0,
        pd.DataFrame({"col1": [1, 2], "col2": [3, 4]}),
        pd.DataFrame(index=range(2)),
        None,
    )
    assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"


def test_lineplot_click():
    click = [{"points": [{"customdata": [0]}]}]
    aux, row, _ = handle_click_events(click, pd.DataFrame(index=range(2)))

    assert all(get_selected(aux) == [True, False])
    assert row == 0


def test_lineplot_hover():
    hover = [{"points": [{"customdata": [1]}]}]
    assert handle_hover_events(hover)[0] == 1



from io import BytesIO


class FinallyCloseBytesIO:
    def __init__(self):
        self.bytes = NoCloseBytesIO()

    def __enter__(self):
        return self.bytes

    def __exit__(self, exc_type, exc_value, traceback):
        super(NoCloseBytesIO, self.bytes).close()


class NoCloseBytesIO(BytesIO):
    def close(self):
        pass



import uuid

import dash
import dash_mantine_components as dmc
from dash import ALL, Input, Output, State, ctx, dcc, html
from dash_extensions.enrich import CycleBreakerInput

from xiplot.plots.barplot import Barplot
from xiplot.plots.boxplot import Boxplot
from xiplot.plots.distplot import Distplot
from xiplot.plots.heatmap import Heatmap
from xiplot.plots.histogram import Histogram
from xiplot.plots.lineplot import Lineplot
from xiplot.plots.scatterplot import Scatterplot
from xiplot.plots.smiles import Smiles
from xiplot.plots.table import Table
from xiplot.tabs import Tab
from xiplot.tabs.plugins import get_plugins_cached
from xiplot.utils import generate_id
from xiplot.utils.auxiliary import merge_df_aux
from xiplot.utils.components import DeleteButton, FlexRow
from xiplot.utils.layouts import layout_wrapper


class Plots(Tab):
    @staticmethod
    def get_plot_types():
        return {
            p.name(): p
            for p in [
                Scatterplot,
                Lineplot,
                Histogram,
                Distplot,
                Barplot,
                Boxplot,
                Heatmap,
                Table,
                Smiles,
            ]
            + [plot for (_, _, plot) in get_plugins_cached("plot")]
        }

    @staticmethod
    def register_callbacks(app, df_from_store, df_to_store):
        for plot_name, plot_type in Plots.get_plot_types().items():
            plot_type.register_callbacks(
                app, df_from_store=df_from_store, df_to_store=df_to_store
            )

        @app.callback(
            Output("plots-tab-settings-session", "children"),
            Output("plots", "children"),
            Output("metadata_store", "data"),
            Output("plots-tab-notify-container", "children"),
            Input("new_plot-button", "n_clicks"),
            Input(generate_id(DeleteButton, ALL), "n_clicks"),
            State("plots", "children"),
            State("plot_type", "value"),
            State("data_frame_store", "data"),
            State("auxiliary_store", "data"),
            CycleBreakerInput("metadata_store", "data"),
            State("plots-tab-settings-session", "children"),
        )
        def add_new_plot(
            n_clicks,
            deletion,
            children,
            plot_type,
            df,
            aux,
            meta,
            last_meta_session,
        ):
            plot_types = Plots.get_plot_types()

            if meta is None:
                if ctx.triggered_id == "new_plot-button":
                    return (
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message="You have not yet loaded any data file.",
                            action="show",
                            autoClose=10000,
                        ),
                    )
                else:
                    return (
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                    )

            if not children:
                children = []

            if ctx.triggered_id in [
                "new_plot-button",
                "metadata_store_data_breaker",
            ]:
                if df is None:
                    return (
                        meta["session"],
                        dash.no_update,
                        dash.no_update,
                        dmc.Notification(
                            id=str(uuid.uuid4()),
                            color="yellow",
                            title="Warning",
                            message="You have not yet loaded any data file.",
                            action="show",
                            autoClose=10000,
                        ),
                    )

                if ctx.triggered_id == "new_plot-button":
                    if not plot_type:
                        return (
                            meta["session"],
                            dash.no_update,
                            dash.no_update,
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message="You have not selected any plot type.",
                                action="show",
                                autoClose=10000,
                            ),
                        )

                    plots = {str(uuid.uuid4()): dict(type=plot_type)}
                elif meta["session"] == last_meta_session:
                    return (
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                        dash.no_update,
                    )
                else:
                    import jsonschema

                    children = []
                    try:
                        jsonschema.validate(
                            instance=meta,
                            schema=dict(
                                type="object",
                                properties=dict(
                                    plots=dict(
                                        type="object",
                                        patternProperties={
                                            r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$": dict(  # noqa: 501
                                                type="object",
                                                properties=dict(
                                                    type=dict(type="string")
                                                ),
                                                required=["type"],
                                            ),
                                        },
                                        additionalProperties=False,
                                    ),
                                ),
                                required=["plots"],
                            ),
                        )
                    except jsonschema.exceptions.ValidationError as err:
                        return (
                            meta["session"],
                            dash.no_update,
                            dash.no_update,
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message=(
                                    "Invalid plots configuration at"
                                    f" meta{err.json_path[1:]}: {err.message}."
                                ),
                                action="show",
                                autoClose=10000,
                            ),
                        )

                    plots = meta["plots"]

                # read df from store
                df = merge_df_aux(df_from_store(df), aux)

                columns = df.columns.to_list()

                notifications = []

                import jsonschema

                for index, config in plots.items():
                    plot_type = config["type"]
                    config = {k: v for k, v in config.items() if k != "type"}

                    if plot_type not in plot_types:
                        notifications.append(
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message=f"Unknown plot type: {plot_type}",
                                action="show",
                                autoClose=10000,
                            )
                        )
                        continue

                    try:
                        layout = plot_types[plot_type].create_new_layout(
                            index,
                            df,
                            columns,
                            config=config,
                        )

                        children.append(layout)
                    except jsonschema.exceptions.ValidationError as err:
                        notifications.append(
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message=(
                                    "Invalid configuration for a"
                                    f" {plot_type} at"
                                    f" meta.plots.{index}{err.json_path[1:]}:"
                                    f" {err.message}"
                                ),
                                action="show",
                                autoClose=10000,
                            )
                        )
                    except ImportError as err:
                        raise err
                    except Exception as err:
                        notifications.append(
                            dmc.Notification(
                                id=str(uuid.uuid4()),
                                color="yellow",
                                title="Warning",
                                message=(
                                    f"Failed to create a {plot_type}: {err}."
                                ),
                                action="show",
                                autoClose=10000,
                            )
                        )

                if ctx.triggered_id == "new_plot-button":
                    return (
                        meta["session"],
                        children,
                        dash.no_update,
                        notifications,
                    )
                else:
                    meta["plots"].clear()
                    return meta["session"], children, meta, notifications

            deletion_id = ctx.triggered_id["index"]

            children = [
                chart
                for chart in children
                if chart["props"]["id"]["index"] != deletion_id
            ]

            return meta["session"], children, dash.no_update, None

    @staticmethod
    def create_layout():
        plots = list(Plots.get_plot_types().keys())
        return html.Div(
            [
                layout_wrapper(
                    component=FlexRow(
                        dcc.Dropdown(
                            plots, plots[0], id="plot_type", clearable=False
                        ),
                        html.Button(
                            "Add", id="new_plot-button", className="button"
                        ),
                    ),
                    title="Select a plot type",
                ),
            ],
            id="control_plots_content-container",
        )

    @staticmethod
    def create_layout_globals():
        return html.Div(
            [
                html.Div(
                    id="plots-tab-notify-container", style={"display": "none"}
                ),
                html.Div(
                    id="plots-tab-settings-session", style={"display": "none"}
                ),
            ],
            style={"display": "none"},
        )



import time

import dash
import pandas as pd
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from tests.util_test import render_plot, start_server
from xiplot.plots.barplot import Barplot

render = Barplot.register_callbacks(
    dash.Dash(__name__), lambda x: x, lambda x: x
)[0]


def test_teba001_render_barplot(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Barplot")

    plot = driver.find_element(By.CLASS_NAME, "dash-graph")

    assert "barplot" in plot.get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_teba002_change_axis_value(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Barplot")

    driver.find_element(By.CLASS_NAME, "dd-double-left").click()

    x = driver.find_element(
        By.XPATH,
        (
            "//div[@class='dd-double-left']"
            "/div[2]/div[1]/div[1]/div[1]/div[2]/input"
        ),
    )

    x.send_keys("model-year")
    x.send_keys(Keys.RETURN)
    time.sleep(0.5)

    assert "model-year" in driver.find_element(By.CLASS_NAME, "xtitle").text
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_teba003_set_cluster(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Barplot")

    # Run clustering
    driver.find_element(By.XPATH, "//div[@id='control-tabs']/div[3]").click()
    feature_dd = driver.find_element(By.ID, "cluster_feature")
    feature_dd.find_element(By.TAG_NAME, "input").send_keys("PCA")
    time.sleep(0.1)
    # The headless driver uses some wierd window size so that the dropdown
    # obscures the button. This is why we have cannot just use `click` here:
    driver.execute_script(
        "arguments[0].click();",
        driver.find_element(By.ID, "add_by_keyword-button"),
    )
    time.sleep(0.1)
    driver.find_element(By.ID, "cluster-button").click()
    time.sleep(0.5)

    # Use clusters
    inp = driver.find_element(
        By.CSS_SELECTOR, ".dd-single.cluster-comparison input"
    )
    inp.send_keys("2")
    inp.send_keys(Keys.RETURN)
    time.sleep(0.5)
    try:
        cluster_val = driver.find_element(
            By.CSS_SELECTOR, ".dd-single.cluster-comparison"
        ).find_element(By.CSS_SELECTOR, ".Select-value")
    except Exception:
        # Sometimes the GitHub test is too slow to find ".Select-value"
        cluster_val = driver.find_element(
            By.CSS_SELECTOR, ".dd-single.cluster-comparison"
        )
    assert "Cluster #2" in cluster_val.get_attribute("innerHTML")

    assert dash_duo.get_logs() == [], "browser console should contain no error"
    driver.close()


def test_teba004_set_order(dash_duo):
    driver = start_server(dash_duo)
    render_plot(dash_duo, driver, "Barplot")

    comparison_order_dd = driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[5]/div[2]",
    )
    comparison_order_dd.click()

    dropdown_input = driver.find_element(
        By.XPATH,
        (
            "//div[@class='plots']/div[4]/div[2]/div[1]/div[1]/div[1]/"
            "div[@class='Select-input']/input"
        ),
    )
    dropdown_input.send_keys("total", Keys.RETURN)
    time.sleep(0.5)

    assert "total" in driver.find_element(
        By.XPATH,
        "//div[@class='plots']/div[4]/div[2]/div[1]/div[1]",
    ).get_attribute("outerHTML")
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    driver.close()


def test_create_barplot():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    output = render("col1", "col2", ["all"], "reldiff", 1, df, pd.DataFrame())
    fig = output[0]
    assert str(type(fig)) == "<class 'plotly.graph_objs._figure.Figure'>"



from collections import defaultdict

import plotly.graph_objects as go
import plotly.io as pio
from dash import Dash, Input, Output, State, dcc, html

from xiplot.tabs import Tab
from xiplot.tabs.plugins import (
    get_all_loaded_plugins_cached,
    is_dynamic_plugin_loading_supported,
)
from xiplot.utils.components import FlexRow
from xiplot.utils.layouts import layout_wrapper


class Settings(Tab):
    @staticmethod
    def register_callbacks(app: Dash, df_from_store, df_to_store):
        pio.templates["xiplot_light"] = go.layout.Template(
            layout={
                "paper_bgcolor": "rgba(255,255,255,0)",
                "margin": dict(l=10, r=10, t=10, b=10, autoexpand=True),
                "uirevision": True,
            }
        )
        pio.templates["xiplot_dark"] = go.layout.Template(
            layout={
                "paper_bgcolor": "rgba(0,0,0,0)",
                "margin": dict(l=10, r=10, t=10, b=10, autoexpand=True),
                "uirevision": True,
            }
        )
        pio.templates.default = "plotly_white+xiplot_light"

        app.clientside_callback(
            """
function toggleLightDarkMode(clicks, data) {
    if (clicks != undefined) {
        data = !data
    }
    if (data) {
        document.documentElement.setAttribute("data-theme", "dark")
        return ['Light mode', data, "plotly_dark+xiplot_dark"]
    } else {
        document.documentElement.setAttribute("data-theme", "light")
        return ['Dark mode', data, "plotly_white+xiplot_light"]
    }
}
            """,
            Output("light-dark-toggle", "children"),
            Output("light-dark-toggle-store", "data"),
            Output("plotly-template", "data"),
            Input("light-dark-toggle", "n_clicks"),
            State("light-dark-toggle-store", "data"),
            prevent_initial_call=False,
        )

        app.clientside_callback(
            """
            function changeColSize(cols) {
                document.documentElement.setAttribute("data-cols", cols)
                return ' '
            }
            """,
            Output("settings-tab-dummy", "children"),
            Input("settings-column-size", "value"),
            prevent_initial_call=False,
        )

        app.clientside_callback(
            """
            function changePlotHeight(value) {
                document.documentElement.setAttribute("plot-height", value)
                return ' '
            }
            """,
            Output("settings-tab-dummy", "children"),
            Input("settings-plot-height", "value"),
            prevent_initial_call=False,
        )

    @staticmethod
    def create_layout():
        return FlexRow(
            *[
                layout_wrapper(
                    component=html.Button(
                        "Dark mode",
                        id="light-dark-toggle",
                        className="light-dark-toggle button",
                    ),
                    title="Colour scheme",
                ),
                html.Span(" ", id="settings-tab-dummy"),
                layout_wrapper(
                    component=dcc.Dropdown(
                        ["1", "2", "3", "4", "5"],
                        "3",
                        clearable=False,
                        persistence="true",
                        id="settings-column-size",
                    ),
                    title="Maximum number of columns",
                ),
                html.Span(" "),
                layout_wrapper(
                    component=dcc.Slider(
                        min=350,
                        max=650,
                        step=100,
                        marks={i: f"{i}" for i in range(350, 651, 100)},
                        value=450,
                        id="settings-plot-height",
                        persistence="true",
                    ),
                    style={"width": "12rem"},
                    title="Plot height",
                ),
            ]
            + (
                []
                if is_dynamic_plugin_loading_supported()
                else [
                    html.Span(" "),
                    layout_wrapper(
                        component=FlexRow(
                            dcc.Dropdown(
                                get_installed_plugin_options(),
                                id="installed_plugins",
                                clearable=False,
                                searchable=False,
                                placeholder=(
                                    "Check the list of installed plugins"
                                ),
                            ),
                        ),
                        title="Installed Plugins",
                        css_class="dash-dropdown",
                    ),
                ]
            ),
            id="control-settings-container",
            style={"alignItems": "start"},
        )

    @staticmethod
    def create_layout_globals():
        globals = [
            # Store the dark/light state across page reloads
            dcc.Store(
                id="light-dark-toggle-store", data=False, storage_type="local"
            ),
            dcc.Store(id="plotly-template", data=None),
        ]
        return html.Div(globals)


def get_installed_plugin_options():
    plugins = defaultdict(set)

    for (
        kind,
        name,
        path,
        plugin,
    ) in get_all_loaded_plugins_cached():
        plugins[path.split(":")[0].split(".")[0]].add(kind)

    plugin_options = []

    for plugin, kinds in plugins.items():
        if len(kinds) > 0:
            kinds = f": {', '.join(kinds)}"
        else:
            kinds = ""

        plugin_options.append(
            {
                "label": f"{plugin}{kinds}",
                "value": plugin,
                "disabled": True,
            }
        )

    return plugin_options



from typing import Any, Dict, Optional

from dash import ALL, ALLSMALLER, MATCH


def generate_id(
    cls: type, index: Optional[Any] = None, subtype: Optional[str] = None
) -> Dict[str, Any]:
    """Generate id:s for (dash) objects.

    Since the type of the id is based on an actual type there is less chance
    for overloading and better refactoring resistance.
    Furthermore, using this function instead of a handcrafted dict leads to
    fewer typos.

    Args:
        cls: The type for which the id is generated, e.g., a subclass of
            `xiplot.plots.APlot`.
        index: The index of the id. Defaults to None.
        subtype: If multiple id:s are generated for the same class (e.g.
            children of a `html.Div`) then this can be used for
            differentiation. Defaults to None.

    Returns:
        An id to be used with Dash.
    """
    if cls in [MATCH, ALL, ALLSMALLER]:
        classtype = cls
    elif subtype is None:
        classtype = f"{cls.__module__}_{cls.__qualname__}".replace(".", "_")
    else:
        classtype = f"{cls.__module__}_{cls.__qualname__}_{subtype}".replace(
            ".", "_"
        )
    if index is None:
        return {"type": classtype}
    else:
        return {"type": classtype, "index": index}
